#include "include/system.h"

#define VIDEO_SEGMENT 0xB8000

uint16 *textmemptr;
uint8 attrib = 0x07;
uint8 csr_x = 0;
uint8 csr_y = 0;

/* Scrolls the screen */
void scroll(void)
{
  uint16 blank;
  uint16 temp;

  /* A blank is defined as a space... we need to give it
  *  backcolor too */
  blank = 0x20 | (attrib << 8);

  /* Row 25 is the end, this means we need to scroll up */
  if(csr_y >= 25)
  {
    /* Move the current text chunk that makes up the screen
     * back in the buffer by a line */
    temp = csr_y - 25 + 1;
    memcpyw (textmemptr, textmemptr + temp * 80, (25 - temp) * 80 * 2);

    /* Finally, we set the chunk of memory that occupies
     * the last line of text to our 'blank' character */
    memsetw (textmemptr + (25 - temp) * 80, blank, 80);
    csr_y = 25 - 1;
  }
}

/* Updates the hardware cursor: the little blinking line
*  on the screen under the last character pressed! */
void move_csr(void)
{
  uint16 offset;

  /* The equation for finding the index in a linear
  *  chunk of memory can be represented by:
  *  Index = [(y * width) + x] */
  offset = csr_y * 80 + csr_x;

  /* This sends a command to indicies 14 and 15 in the
  *  CRT Control Register of the VGA controller. These
  *  are the high and low bytes of the index that show
  *  where the hardware cursor is to be 'blinking'. To
  *  learn more, you should look up some VGA specific
  *  programming documents. A great start to graphics:
  *  http://www.brackeen.com/home/vga */
  outportb(0x3D4, 14);
  outportb(0x3D5, offset >> 8);
  outportb(0x3D4, 15);
  outportb(0x3D5, offset);
}

void cls()
{
  /* Fill screen with blank spaces (with current attribute) */	
	uint16 blank = 0x20 | (attrib << 8);
  memsetw ((uint16*) textmemptr, blank, 80 * 25);

  /* Move hardware cursor to 0,0. */
  csr_x = 0;
  csr_y = 0;
  move_csr();
}


/* Puts a single character on the screen */
void putch(char c)
{
  uint16 *where;
  uint16 att = attrib << 8;

  /* Handle a backspace, by moving the cursor back one space */
  if(c == 0x08)
  {
      if(csr_x != 0) csr_x--;
  }
  /* Handles a tab by incrementing the cursor's x, but only
  *  to a point that will make it divisible by 8 */
  else if(c == 0x09)
  {
      csr_x = (csr_x + 8) & ~(8 - 1);
  }
  /* Handles a 'Carriage Return', which simply brings the
  *  cursor back to the margin */
  else if(c == '\r')
  {
      csr_x = 0;
  }
  /* We handle our newlines the way DOS and the BIOS do: we
  *  treat it as if a 'CR' was also there, so we bring the
  *  cursor to the margin and we increment the 'y' value */
  else if(c == '\n')
  {
      csr_x = 0;
      csr_y++;
  }
  /* Any character greater than and including a space, is a
  *  printable character. The equation for finding the index
  *  in a linear chunk of memory can be represented by:
  *  Index = [(y * width) + x] */
  else if(c >= ' ')
  {
      where = textmemptr + (csr_y * 80 + csr_x);
      *where = c | att;	/* Character AND attributes: color */
      csr_x++;
  }

  /* If the cursor has reached the edge of the screen's width, we
  *  insert a new line in there */
  if(csr_x >= 80)
  {
      csr_x = 0;
      csr_y++;
  }

  /* Scroll the screen if needed, and finally move the cursor */
  scroll();
  move_csr();
}

/* Uses the above routine to output a string... */
void puts(char *text)
{
	uint32 len = strlen(text);
  uint32 i;

  for (i = 0; i < len; i++)
  {
    putch(text[i]);
  }
}

/* Sets the forecolor and backcolor that we will use */
void settextcolor(uint8 forecolor, uint8 backcolor)
{
  /* Top 4 bytes are the background, bottom 4 bytes
   * are the foreground color */
  attrib = (backcolor << 4) | (forecolor & 0x0F);
}

void init_video()
{
	char* str = "Hello, world";
	textmemptr = (uint16*) 0xb8000;
	
	
	cls();
	puts(str);
}